#!/usr/bin/perl 

package PasswordCrypt;

use Digest::MD5;
use DomainAdminConfig;

my %supported_methods = (
    'crypt' => 1,
    'crypt-md5' => 1,
    'md5' => 1,
    'plaintext' => 1
);

#
# This is the method to use for hashes that don't start with {method}.  As
# opposed to DomainAdminConfig::default_hash_method which specifies the
# method to use for newly created hashes
#
my $old_default_method = "md5";

my $new_default_method = $DomainAdminConfig::default_hash_method ? 
	$DomainAdminConfig::default_hash_method : $old_default_method;

my $itoa64 = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
my @itoa64 = split //, $itoa64;

#
# is_supported:
#
sub is_supported {
    my $method = shift;
    return $supported_methods{$method} ? 1 : 0;
}

#
# methods: returns list of supported methods, default first
#
sub methods {
    # This highly suspect code ensures that the default appears first.
    return sort { ($b eq $new_default_method) <=> ($a eq $new_default_method) } keys(%supported_methods);
}

#
# authenticate:
#
sub authenticate {
    my ($password, $hash) = @_;
    my $method;
    if ($hash =~ m/^\{([^}]+)\}(.*)$/) {
        $method = $1; 
        $hash = $2;
    } else {
        $method = $old_default_method;
    }

    if ($method eq 'crypt') {
        return($hash and (crypt($password, $hash) eq $hash));
    } elsif ($method eq 'plaintext') {
        return($hash and ($password eq $hash));
    } elsif ($method eq 'md5') {
        return($hash and (Digest::MD5::md5_hex($password) eq $hash));
    } elsif ($method eq 'crypt-md5') {
        print STDERR "$password  $hash  ".crypt_md5($password,$hash)."\n";
        return($hash and (crypt_md5($password, $hash) eq $hash));
    } else {
        return(undef);
    }
}

#
# hash: method can be either an encryption method or an old hash
#
sub hash {
    my ($password, $method) = @_;
    if (!$method) { 
        $method = $new_default_method;
    } 

    my $hash;

    if ($method eq 'crypt') {
        #
        # For crypt, we use standard two digit salts, not "smart" 
        # "we want MD5" salts.
        #
        my $salt = $itoa64[rand(64)] . $itoa64[rand(64)];
        $hash = "{crypt}" . crypt($password,$salt);
    } elsif ($method eq 'md5') {
        $hash = "{md5}".Digest::MD5::md5_hex($password);
    } elsif ($method eq 'crypt-md5') {
        my $salt;
        for (my $i = 0; $i < 7; $i++) {
            $salt .= $itoa64[rand(64)];
        }
        $salt = '$1$' . $salt . '$';
        $hash = "{crypt-md5}".crypt_md5($password,$salt);
    } elsif ($method eq 'plaintext') {
        $hash = "{plaintext}".$password;
    }
    return $hash;
}
 
#
# crypt_md5.pl:
# Implementation of Poul-Henning Kemp's crypt_md5 routine in perl.
#
# Copyright (c) 2001 Chris Lightfoot. All rights reserved.
#
#

sub to64 ($$) {
    my ($num, $len) = @_;
    my $ret = "";
    while (--$len >= 0) {
        $ret .= substr($itoa64, $num & 0x3f, 1);
        $num >>= 6;
    }
    return $ret;
}

sub crypt_md5 ($$) {
    my ($pw, $sp) = @_;

    # refine the salt

    # skip magic string
    $sp =~ s/^\$1\$//;

    my ($salt) = ($sp =~ /^([^\$]{1,8})\$/);
    return undef unless defined($salt);

    my $ctx = new Digest::MD5;

    # Add password...
    $ctx->add($pw);

    # Now magic string
    $ctx->add('$1$');

    # Raw salt 
    $ctx->add($salt);

    my $ctx1 = new Digest::MD5;

    $ctx1->add($pw);
    $ctx1->add($salt);
    $ctx1->add($pw);

    my $final = $ctx1->digest();

    # `Just as many characters of the MD5(pw, salt, pw)'
    for (my $pl = length($pw); $pl > 0; $pl -= 16) {
        $ctx->add(substr($final, 0, $pl > 16 ? 16 : $pl));
    }

    $final = "\0" x 16;

    # Now `something really weird'.
    #
    # I believe that this bit of crypt_md5 has been mis-implemented, but I am
    # not sure.
    for (my $i = length($pw); $i; $i >>= 1) {
        if ($i & 1) {  
            $ctx->add("\0");
        } else {
            $ctx->add(substr($pw, 0, 1));
        }
    }

    $final = $ctx->digest();


    for (my $i = 0; $i < 1000; ++$i) {
        $ctx1 = undef;
        $ctx1 = new Digest::MD5;
        if ($i & 1) {
            $ctx1->add($pw);
        } else {
            $ctx1->add($final);
        }

        if ($i % 3) {
            $ctx1->add($salt);
        }

        if ($i % 7) {
            $ctx1->add($pw);
        }

        if ($i & 1) {
            $ctx1->add($final);
        } else {
            $ctx1->add($pw);
        }

        $final = $ctx1->digest();
    }

    my @f = unpack("C16", $final);
    $final = "\0" x 16;

    my $ret = '$1$' . $salt . '$'; 

    my $l;

    $l = ($f[ 0]<<16) | ($f[ 6]<<8) | $f[12]; $ret .= to64($l, 4);
    $l = ($f[ 1]<<16) | ($f[ 7]<<8) | $f[13]; $ret .= to64($l, 4);
    $l = ($f[ 2]<<16) | ($f[ 8]<<8) | $f[14]; $ret .= to64($l, 4);
    $l = ($f[ 3]<<16) | ($f[ 9]<<8) | $f[15]; $ret .= to64($l, 4);
    $l = ($f[ 4]<<16) | ($f[10]<<8) | $f[ 5]; $ret .= to64($l, 4);
    $l =                    $f[11]          ; $ret .= to64($l, 2);
}


